; File:		ir-tx-nec-mega88a\main.asm
; Created:	16.10.2023 13:07:23
; Device:	ATmega88A
; Version:	2023-10-18

.include <m88Adef.inc>
.include "macros.asm"

; == General definitions ==
; Set the CPU clock frequency in Hz:
.equ F_CPU = 7_372_800

; == IR definitions ==
; Set the IR address and some commands:
.equ IR_ADDRESS = 0
.equ IR_CMD_POWER = $45
; NEC modulation definitions (560 s timer interrupt counts, decremented by 1),
; values corresponding to https://www.mikrocontroller.net/articles/IRMP#NEC_+_extended_NEC :
.equ IR_PULSE = 0
.equ IR_PAUSE_0 = 0
.equ IR_PAUSE_1 = 2
.equ IR_PAUSE_START_DATA = 7
.equ IR_PULSE_START = 15
.equ IR_PULSE_COUNT = 34		; including start and stop bit
.equ IR_DATA_START = 2			; value of ir_pulseCntr while first data pulse+pause is being sent

; == Hardware definitions ==
; Set switch connection:
.equ SWITCH_PIN = PINC
.equ SWITCH = 0
; Set OC0A = IR output (high active -> LED on at high state):
.equ OC0A_DDR = DDRD
.equ OC0A_PORT = PORTD
.equ OC0A = 6

; == Register definitions ==
; === Auxiliary registers ===
.def zero = r2
.def temp0 = r16
.def temp1 = r17
.def temp2 = r18

; === IR TX registers ===
.def ir_intrCntr = r3
.def ir_pulseCntr = r19			; Rising-edge (high-active) counter: 0 before first falling edge,
								; 1 while start pulse+pause, 2 while first data pulse+pause, ...

; == RAM usage ==
.dseg
.org SRAM_START
ir_data: .byte 4	; The data to be transmitted (address and command):
					; ir_data		address
					; ir_data+1		address one's complement
					; ir_data+2		command
					; ir_data+3		command one's complement

.cseg

.org 0
	rjmp	reset
.org OC1Aaddr					; Timer1 Compare Match A interrupt
	rjmp	ir_intr
.org INT_VECTORS_SIZE

reset:
; Initialize stack pointer:
	ldi		temp0, high(RAMEND)
	xout	SPH, temp0
	ldi		temp0, low(RAMEND)
	xout	SPL, temp0
; Initialize general registers:
	clr		zero
; Initialize IR registers:
	ldi		temp0, IR_ADDRESS
	sts		ir_data+0, temp0
	com		temp0
	sts		ir_data+1, temp0
	clr		ir_intrCntr
	clr		ir_pulseCntr
; Initialize Timer0: 36 kHz carrier, output to PD6=OC0A:
	ldi		temp0, ROUND(F_CPU/(2*36_000))-1
	xout	OCR0A, temp0
	ldi		temp0, (1<<COM0A0) | (1<<WGM01)			; toggle OC0A, CTC mode
	xout	TCCR0A, temp0
; Initialize Timer1: 560 s "tick":
.equ TIMER1_OCR = ROUND(1.0*F_CPU*(5.6e-4))-1		; should be 4,128
	ldi		temp0, high(TIMER1_OCR)
	xout	OCR1AH, temp0
	ldi		temp0, low(TIMER1_OCR)
	xout	OCR1AL, temp0
	ldi		temp0, 1<<WGM12							; CTC mode
	xout	TCCR1A, temp0
	ldi		temp0, 1<<OCIE1A						; enable Compare Match A interrupt
	xout	TIMSK1, temp0
; Initialize IR output:
	sbi		OC0A_DDR, OC0A
; Enable interrupts:
	sei
; Load IR command:
	ldi		temp0, IR_CMD_POWER
; Main loop:
loop:
	sbic	SWITCH_PIN, SWITCH
	rjmp	loop
	rcall	ir_transmit
	rcall	delay1s					; primitive switch debouncing
									; (for testing only)
	rjmp	loop

; == IR transmitter macros ==
.macro ir_carrierOn
	ldi		temp0, 1<<CS00			; no prescaling
	xout	TCCR0B, temp0
	sbi		OC0A_PORT, OC0A			; switch on IR LED
.endmacro

.macro ir_carrierOff
	xout	TCCR0B, zero			; stop timer
	cbi		OC0A_PORT, OC0A			; switch off IR LED
	xout	TCNT0, zero
.endmacro

; == IR transmitter interrupt routine ==
ir_intr:
; Save temp0, temp1, temp2, and SREG:
	push	temp0
	push	temp1
	push	temp2
	xin		temp0, SREG
	push	temp0
; Branch according to interrupt counter:
	tst		ir_intrCntr
	breq	ir_intr_0
	dec		ir_intrCntr
	rjmp	ir_intr_end
ir_intr_0:
; Branch according to pulseCntr:
	cpi		ir_pulseCntr, 0
	breq	ir_intr_1stEdge			; first edge, beginning of start pulse
	cpi		ir_pulseCntr, 1
	breq	ir_intr_start			; either after start pulse or start pause
	cpi		ir_pulseCntr, 34
	brlo	ir_intr_data			; either after data pulse or data pause, index=pulseCntr-2
; After stop pulse. Switch off carrier and interrupt timer:
	ir_carrierOff
	xout	TCCR1B, zero			; stop timer
	xout	TCNT1H, zero
	xout	TCNT1L, zero
	clr		ir_pulseCntr
	rjmp	ir_intr_end
ir_intr_1stEdge:
; Before start pulse.
	ir_carrierOn
	ldi		temp0, IR_PULSE_START
	mov		ir_intrCntr, temp0
	rjmp	ir_intr_inc
ir_intr_start:
	xin		temp0, TCCR0B
	sbrc	temp0, CS00				; skip if carrier is inactive
	rjmp	ir_intr_start_1
; After start pause.
	ir_carrierOn
	ldi		temp0, IR_PULSE
	mov		ir_intrCntr, temp0
	rjmp	ir_intr_inc
ir_intr_start_1:
; After start pulse.
	ir_carrierOff
	ldi		temp0, IR_PAUSE_START_DATA
	mov		ir_intrCntr, temp0
	rjmp	ir_intr_end
ir_intr_data:
	xin		temp0, TCCR0B
	sbrc	temp0, CS00				; skip if carrier is inactive
	rjmp	ir_intr_data_1
; After data pause.
	ir_carrierOn
	ldi		temp0, IR_PULSE
	mov		ir_intrCntr, temp0
	rjmp	ir_intr_inc
ir_intr_data_1:
; After data pulse.
	ir_carrierOff
; Load data byte:
	mov		temp0, ir_pulseCntr
	subi	temp0, IR_DATA_START		; subtract offset
	mov		temp1, temp0
	lsr		temp0						; Divide ...
	lsr		temp0
	lsr		temp0						; ... by 8.
	ldiz	ir_data
	addz	temp0
	ld		temp0, Z
; Prepare the necessary bit mask:
	andi	temp1, $07					; temp1 <- temp1 % 8
	ldi		temp2, 1					; becomes the bit mask
	tst		temp1
ir_intr_data_1_loop:
	breq	ir_intr_data_1_afterLoop
	lsl		temp2
	dec		temp1
	rjmp	ir_intr_data_1_loop
ir_intr_data_1_afterLoop:
; Get the data bit value:
	and		temp0, temp2
	brne	ir_intr_data_1_1
; Data bit == 0.
	ldi		temp0, IR_PAUSE_0
	rjmp	ir_intr_data_2
ir_intr_data_1_1:
; Data bit == 1.
	ldi		temp0, IR_PAUSE_1
ir_intr_data_2:
	mov		ir_intrCntr, temp0
	rjmp	ir_intr_end
ir_intr_inc:
	inc		ir_pulseCntr
ir_intr_end:
; Restore SREG, temp2, temp1, and temp0:
	pop		temp0
	xout	SREG, temp0
	pop		temp2
	pop		temp1
	pop		temp0
	reti

; == Routine delay1s ==
; (for testing only)
delay1s:
	push	temp0
	push	temp1
	push	temp2
	ldi		temp0, F_CPU*10/607/1_000
delay1s_loop0:
	ldi		temp1, 201
delay1s_loop1:
	ldi		temp2, 100
delay1s_loop2:
	dec		temp2
	brne	delay1s_loop2
	dec		temp1
	brne	delay1s_loop1
	dec		temp0
	brne	delay1s_loop0
	pop		temp2
	pop		temp1
	pop		temp0
	ret

; == Routine ir_transmit ==
; Parameter: temp0: command
ir_transmit:
	push	temp0
	sts		ir_data+2, temp0
	com		temp0
	sts		ir_data+3, temp0
; Start Timer1:
	ldi		temp0, 1<<CS10			; no prescaling
	xout	TCCR1B, temp0
	pop		temp0
	ret
